from functools import wraps

import httpx


def memweb(url):
    """
    Remembers input/output of a function by sending it over http to an endpoint.

    Arguments:
        filepath: path to write data to
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            with httpx.Client() as client:
                _ = client.post(url, data={**kwargs, **result})
            return result

        return wrapper

    return decorator



from ._error import NotInstalled
from ._grid import grid, random_grid
from ._base import memlist, memfile, memfunc
from ._util import time_taken

try:
    from memo._http import memweb
except ModuleNotFoundError:
    memweb = NotInstalled("memweb", "web")

try:
    from memo._wandb import memwandb
except ModuleNotFoundError:
    memwandb = NotInstalled("memwandb", "wandb")


__version__ = "0.1.1"
__all__ = [
    "grid",
    "random_grid",
    "memlist",
    "memfile",
    "memfunc",
    "memweb",
    "memwandb",
    "time_taken",
]



from functools import wraps

import wandb


def memwandb(project, **config):
    """
    Remembers input/output of a function by sending it to wandb.

    Warning:
        To use this tool, you may need to install extra dependencies.

        ```
        pip install memo[wandb]
        ```

    Arguments:
        project: name of the wandb project you want to send to
        config: optional extra keyword arguments to send along

    Usage:

    ```python
    import os
    from memo import wemwandb, time_taken

    # You probably want to run with this setting to prevent a
    # whole lot of output suddenly appearing.
    os.environ["WANDB_SILENT"] = "true"

    data = []


    @memwandb(project="birthday")
    @time_taken()
    def birthday_experiment(class_size, n_sim):
        sims = np.random.randint(1, 365 + 1, (n_sim, class_size))
        sort_sims = np.sort(sims, axis=1)
        n_uniq = (sort_sims[:, 1:] != sort_sims[:, :-1]).sum(axis = 1) + 1
        proba = np.mean(n_uniq != class_size)
        return {"est_proba": proba}

    settings = grid(class_size=range(2, 40), n_sim=[1_000, 10_000, 100_000, 1_000_000])
    for setting in settings:
        birthday_experiment(**setting)
    ```
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            import logging

            logger = logging.getLogger("wandb")
            logger.setLevel(logging.ERROR)
            wandb.init(project=project)
            wandb.config = config
            result = func(*args, **kwargs)
            wandb.log({**kwargs, **result})
            return result

        return wrapper

    return decorator



import time
from functools import wraps


def time_taken(minutes=False, rounding=2):
    """
    Adds additional time-based information to output.

    Arguments:
        minutes: log minutes instead of seconds
        rounding: number of decimals to round the timing to

    Example:

    ```python
    import numpy as np

    from memo import memlist, grid, time_taken

    data = []


    @memlist(data=data)
    @time_taken()
    def birthday_experiment(class_size, n_sim):
        sims = np.random.randint(1, 365 + 1, (n_sim, class_size))
        sort_sims = np.sort(sims, axis=1)
        n_uniq = (sort_sims[:, 1:] != sort_sims[:, :-1]).sum(axis = 1) + 1
        proba = np.mean(n_uniq != class_size)
        return {"est_proba": proba}


    settings = grid(class_size=range(5, 50), n_sim=[100, 10_000, 1_000_000])

    for setting in settings:
        birthday_experiment(**setting)
    ```
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            tic = time.time()
            result = func(*args, **kwargs)
            toc = time.time()
            time_total = toc - tic
            if minutes:
                time_total = time_total / 60
            result = {**result, "time_taken": round(time_total, rounding)}
            return result

        return wrapper

    return decorator



import random
import itertools as it


def grid(**kwargs):
    """
    Generates a grid of settings.

    Arguments:
        kwargs: the name of parameter is the key while the values represent items to iterate over

    Example

    ```python
    from memo import grid

    settings = list(grid(a=[1,2], b=[1, 2]))
    expected = [
         {'a': 1, 'b': 1},
         {'a': 1, 'b': 2},
         {'a': 2, 'b': 1},
         {'a': 2, 'b': 2}
    ]
    assert settings == expected

    def calc_sum(a, b):
        return {"c": a + b}

    for setting in settings:
        print(calc_sum(**setting))
    ```
    """
    for c in it.product(*[v for v in kwargs.values()]):
        yield {k: v for k, v in zip(kwargs.keys(), c)}


def random_grid(n=30, **kwargs):
    """
    Generates a random grid settings.

    Arguments:
        kwargs: the name of parameter is the key while the values represent items to iterate over

    Example

    ```python
    from memo import random_grid

    settings = list(random_grid(n=30, a=[1,2], b=[1, 2]))
    assert len(settings) == 30
    ```
    """
    for i in range(n):
        yield {k: random.choice(v) for k, v in kwargs.items()}



from memo import __version__
from setuptools import setup, find_packages

base_packages = []

test_packages = ["flake8>=3.6.0", "pytest>=4.0.2", "numpy>=1.19.4", "exdown>=0.7.1", "rich>=9.2.0", "tqdm>=4.54.0"]

util_packages = [
    "jupyter>=1.0.0",
    "jupyterlab>=0.35.4",
]

docs_packages = [
    "mkdocs>=1.1",
    "mkdocs-material>=4.6.3",
    "mkdocstrings>=0.8.0",
]

dev_packages = util_packages + docs_packages + test_packages

wandb_packages = ["wandb>=0.10.11"]

web_packages = ["httpx>=0.16.1"]

setup(
    name="memo",
    version=__version__,
    packages=find_packages(exclude=["notebooks"]),
    install_requires=base_packages,
    extras_require={
        "web": web_packages,
        "wandb": wandb_packages,
        "test": test_packages,
        "dev": dev_packages,
    },
)



import pathlib

import exdown
import pytest
import textwrap

from memo import memlist, memfunc, memfile, time_taken, grid, random_grid

files = [str(p) for p in pathlib.Path("docs").glob("*.md")] + ["README.md"]
functions = [memlist, memfunc, memfile, time_taken, grid, random_grid]


@pytest.mark.parametrize("fpath", files)
def test_markdown_files(fpath):
    for string, lineno in exdown.extract(fpath, syntax_filter="python"):
        try:
            exec(string, {"__MODULE__": "__main__"})
        except Exception:
            print(f"{fpath} (line {lineno}):\n```\n{string}```")
            raise


@pytest.mark.parametrize("func", functions, ids=lambda d: d.__name__)
def test_docstrings(func, tmp_path):
    with open(f"{tmp_path}/supertmp.md", "w") as handle:
        handle.write(textwrap.dedent(func.__doc__))
    for string, lineno in exdown.extract(
        f"{tmp_path}/supertmp.md", syntax_filter="python"
    ):
        try:
            exec(string, {"__MODULE__": "__main__"})
        except Exception:
            print(f"{func} (line {lineno}):\n```\n{string}```")
            raise
